#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
Monda access functions to retrieve Mini-Secchi disk from PML Geoserver using WFS standard.

To see all available command line options use
 python minisecchi_retrieval_test.py -h

Example use
 python minisecchi_retrieval_test.py -i "2020-01-01" -f "2021-12-31" -t ./secchitest.csv --bbox 50 20 40 30

This will retrieve all records from 2020-2021 within a bounding box spanning 50N20E - 40N30E and save results to secchitest.csv in the current folder.


------------------------------------------------------------------------------
Stefan Simis - stsi@pml.ac.uk - Feb 2022
"""

import os
import sys
import logging
import datetime
import argparse
from monda.minisecchi import access

log = logging.getLogger('secchi-test')
myFormat = '%(asctime)s | %(name)s | %(levelname)s | %(message)s'
formatter = logging.Formatter(myFormat)
logging.basicConfig(level = 'INFO', format = myFormat, stream = sys.stdout)


def parse_args():
    default_start = datetime.datetime.strptime('2022-01-01', '%Y-%m-%d')
    default_end = datetime.datetime.strptime('2022-07-20', '%Y-%m-%d')

    parser = argparse.ArgumentParser()

    parser.add_argument('-i','--initial_day', required = False, type = lambda s: datetime.datetime.strptime(s, '%Y-%m-%d'), default = default_start, help = "Intial day in format ['yyyy-mm-dd]'] ")
    parser.add_argument('-f','--final_day', required = False, type = lambda s: datetime.datetime.strptime(s, '%Y-%m-%d'), default = default_end, help = "Intial day in format ['yyyy-mm-dd]'] ")
    parser.add_argument('-t','--target_file', required = False, type = str, default = None, help = "Target file path to store results in csv format")
    parser.add_argument('-b','--bbox', required = False, type = float, nargs='+', default = None, help = "Restrict query to bounding box format [corner1lat corner1lon corner2lat corner2lon]")
    args = parser.parse_args()

    return args


def run_example(start_date, end_date, geoserver_layer='rsg:miniscecchi_public_view', target=None, bbox=None):

    response = access.get_wfs(count=1000, timewindow=(start_date, end_date), layer=geoserver_layer, bbox=bbox)
    data = response['result']

    try:
        datafields = data[0].keys()
    except IndexError:
        print("no data found for time and region specified")
    else:
        header = ",".join([k for k in datafields])

    if target is not None:
        # attempt to write data to target file
        if os.path.isfile(target):
            log.warning(f"Existing data file will be overwritten")

        with open(target, 'w') as outfile:
            outfile.write(header + '\n')
            for i, record in enumerate(data):
                record_str = ",".join([str(v) for v in record.values()])
                if i < len(data):
                    outfile.write(record_str + '\n')
                else:
                    outfile.write(record_str)


if __name__ == '__main__':
    args = parse_args()
    run_example(args.initial_day, args.final_day, geoserver_layer='rsg:minisecchi_public_view', target=args.target_file, bbox=args.bbox)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
Monda access functions to retrieve HSP data from PML Geoserver using WFS standard.
------------------------------------------------------------------------------
Tom Jordan - tjor@pml.ac.uk - May 2022
Stefan Simis - stsi@pml.ac.uk - May 2022

"""

import logging
import sys
import numpy as np
import datetime
import json
import urllib.request
import urllib.parse

log = logging.getLogger('hsp-downloader')
myFormat = '%(asctime)s | %(name)s | %(levelname)s | %(message)s'
formatter = logging.Formatter(myFormat)
logging.basicConfig(level = 'INFO', format = myFormat, stream = sys.stdout)


def get_wfs(count=1000, sensor=None, timewindow=None, layer='rsg:hsp_public_view_full', bbox=None):
    """
    Get features from geoserver layer, parse json response into python friendly format

    : int count:        Number of features to retrieve per request. There is typically a request feature limit for each geoserver layer. This enables paging the response. (default 1000)
    : str sensor:       Optionally specify a sensor id (default None)
    : tuple timewindow: Specify a time window as a tuple of datetime or iso formatted strings. The default None will request data from the last 24h
    : str layer:        Specify the geoserver layer. For a list of available layers see https://rsg.pml.ac.uk/geoserver/. (default rsg:hsp_public_view_full)

    """
    if isinstance(timewindow, tuple):
        if (isinstance(timewindow[0], str)) and (isinstance(timewindow[1], str)):
            try:
                time_start = datetime.datetime.strftime(timewindow[0], '%Y-%m-%dT%H:%M:%SZ')
                time_end   = datetime.datetime.strftime(timewindow[1], '%Y-%m-%dT%H:%M:%SZ')
            except:
                log.error("timewindow should be a tuple of type datetime.datetime or iso-formatted string i.e. 'YYYY-mm-ddTHH:MM:SSZ'")
                return None
        elif (isinstance(timewindow[0], datetime.datetime)) and (isinstance(timewindow[1], datetime.datetime)):
            time_start = timewindow[0]
            time_end = timewindow[1]
        else:
            log.error("timewindow should be a tuple of type datetime.datetime or iso-formatted string i.e. 'YYYY-mm-ddTHH:MM:SSZ'")
            return None
    elif timewindow is None:
        # default to the last 24H
        time_start = datetime.datetime.now() - datetime.timedelta(days=1)
        time_end = datetime.datetime.now()

    # build a CQL filter to get a time slice
    cql = ''

    cql += f"spectral_record_end_time between {time_start.isoformat()} AND {time_end.isoformat()}"

    if sensor is not None:
        if len(cql)>0:
            cql += f" AND sensor_id='{sensor}'"

    if bbox is not None:
        if not len(bbox) == 4:
           log.error("Bounding box expects a tuple of four values (two corner coordinate pairs)")
        if len(cql)>0:
            cql += " AND "
        cql += f"""BBOX(spectral_record_location, {",".join([str(b) for b in bbox])})"""

    # sanity-check the request count
    layer_limit = 10000  # geoserver layer is limited to 10,000 items per request
    if count > layer_limit:
        log.warning(f"Feature request count adjusted to {layer_limit} as supported by this server")
        count = layer_limit

    # build base URL
    url = 'https://rsg.pml.ac.uk/geoserver/rsg/wfs'
    req = 'GetFeature'
    version = '2.0.0'
    service = 'WFS'
    typeName = layer
    srs = 'EPSG:4326'
    fmt = 'json'
    wfs_url = f"{url}?request={req}&version={version}&service={service}&typeName={typeName}"
    wfs_url += f"&count={count}&srsname={srs}&outputFormat={fmt}&sortBy=spectral_record_end_time+A"
    if cql:
        wfs_url += f"&CQL_FILTER={urllib.parse.quote(cql)}"

    # prepare to page results by first getting number of hits on query, without returning features
    try:
        hits_request = wfs_url + "&resultType=hits"
        log.info(f"Pre-paging request: {hits_request}")
        resp = urllib.request.urlopen(hits_request).read()
    except urllib.request.HTTPError as err:
        log.error(f"""WFS Server <a href="{url}">{url}</a> error""")
        log.exception(err)
        return None

    hitsdata = str(resp)
    if 'numberMatched' in hitsdata:
        hits = int(hitsdata.split('numberMatched="')[1].split('"')[0])
        log.info(f"{hits} features matched")
    else:
        log.error(f"No features match this request. The server response was {resp}")
        return None

    if count < hits:
        # the number of features matching the request exceeds the volume we have set to retreive in one go
        npages = int(np.ceil(hits/count))
        log.info(f"Paging the request: {npages} pages")
    else:
        npages = 1

    paged_result = []
    for page in range(npages):
        startIndex = page * count
        paged_wfs_url = wfs_url + f"&startIndex={startIndex}"
        log.debug(paged_wfs_url)

        try:
            resp = urllib.request.urlopen(paged_wfs_url).read()

        except urllib.request.HTTPError as err:
            log.error(f"""WFS Server <a href="{url}">{url}</a> error""")
            log.exception(err)
            return None

        try:
            data = json.loads(resp)
        except json.JSONDecodeError as err:
            log.error(f"Could not decode JSON response")
            log.exception(err)
            return None

        # flatten the response and deal with special cases
        features = [None]*len(data['features'])
        for i, datum in enumerate(data['features']):
            feature = datum['properties']
            feature['lon'] = datum['geometry']['coordinates'][0]
            feature['lat'] = datum['geometry']['coordinates'][1]
            for key, val in feature.items():
                if isinstance(val, str):
                    if val.count(',') > 0:  # convert comma-separated strings to numpy arrays
                        feature[key] = np.array([float(w) for w in val.split(',')])
                if ('date' in key) or ('time' in key):  # try to convert into timestamp
                    if key in ['timezone']:
                        continue
                    if feature[key] is None:
                        continue
                    try:
                        feature[key] = datetime.datetime.strptime(val, '%Y-%m-%dT%H:%M:%S.%fZ')
                    except ValueError:
                        try:
                            feature[key] = datetime.datetime.strptime(val, '%Y-%m-%dT%H:%M:%SZ')
                        except ValueError:
                            log.error(f"Error parsing date or time field: {val}")
                    except TypeError:
                        pass
                    except Exception as err:
                            log.exception(err)
            features[i] = feature

        log.info(f"Page {page}, starting at count {startIndex}: {len(features)} features")
        paged_result = paged_result + features

    return {'result':paged_result, 'length':len(paged_result)}



from . import WISP
from . import sorad

# import importlib
# import pkgutil
#
# def import_submodules(package, recursive=True):
#     """ Import all submodules of a module, recursively, including subpackages
#
#     :param package: package (name or actual module)
#     :type package: str | module
#     :rtype: dict[str, types.ModuleType]
#     """
#     if isinstance(package, str):
#         package = importlib.import_module(package)
#     results = {}
#     for loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
#         full_name = package.__name__ + '.' + name
#         results[full_name] = importlib.import_module(full_name)
#         if recursive and is_pkg:
#             results.update(import_submodules(full_name))
#     return results





# -*- coding: utf-8 -*-
"""

Monda example script to retrieve HSP data from PML Geoserver using WFS standard,

---------------------------------------------------------------------------------

To see all command line options use:
    python hsp_retrieval_test.py -h

-------------------------------------------------------------------------------

Stefan Simis - stsi@pml.ac.uk - May 2022
"""

import sys
import os
import numpy as np
from monda.hsp import access
import datetime
import logging
import argparse

log = logging.getLogger('sorad-test')
myFormat = '%(asctime)s | %(name)s | %(levelname)s | %(message)s'
formatter = logging.Formatter(myFormat)
logging.basicConfig(level = 'INFO', format = myFormat, stream = sys.stdout)

def run_example(sensor_id = None,
                start_time = datetime.datetime(2021,10,21,12,0,0),
                end_time   = datetime.datetime(2021,10,21,14,0,0),
                bbox = None):

    """
    Download data from a HSP instrument

    sensor_id: the serial number of a So-Rad platform, or None
    start_time:  date/time to start collecting data from
    end_time:    date/time to start collecting data from
    bbox:        Bounding box corner coordinates (lat,lon,lat,lon)
    """

    log.info(f"Request timeframe {start_time.isoformat()} - {end_time.isoformat()}")

    layer = 'rsg:hsp_public_view_full'

    response = access.get_wfs(sensor = sensor_id,
                              timewindow = (start_time, end_time),
                              layer=layer, bbox=bbox)


    log.info(f"{response['length']} features received.")

    if response['length'] == 0:
        return None



    # show first record
    for key, val in response['result'][0].items():
        log.info(f"{key}: {val}")


    return response


def parse_args():
    """Interpret command line arguments"""
    parser = argparse.ArgumentParser()
    parser.add_argument('-s','--sensor',      required = False, type = str, default = None, help = "Instrument serial number")
    parser.add_argument('-i','--start_time',  required = False, type = lambda s: datetime.datetime.strptime(s, '%Y-%m-%d %H:%M:%S'),
                                              default = datetime.datetime(2021,10,21,12,0,0),
                                              help = "Initial UTC date/time in format 'YYYY-mm-dd HH:MM:SS'")
    parser.add_argument('-e','--end_time',    required = False, type = lambda s: datetime.datetime.strptime(s, '%Y-%m-%d %H:%M:%S'),
                                              default = datetime.datetime(2021,10,21,14,0,0),
                                              help = "Final UTC date/time in format 'YYYY-mm-dd HH:MM:SS'")
    parser.add_argument('-b','--bbox',        required = False, type = float, nargs='+', default = None, help = "Restrict query to bounding box format [corner1lat corner1lon corner2lat corner2lon]")

    args = parser.parse_args()

    return args


if __name__ == '__main__':

    args = parse_args()

    response = run_example(args.sensor, args.start_time, args.end_time, args.bbox)



# -*- coding: utf-8 -*-
"""

Monda example script to retrieve So-Rad Rrs and (ir)radiance spectra from PML Geoserver using WFS standard,
filter data using Quality Control principles, plot results, save Rrs spectra and some metadata (including QC masks).
The data and plots will be written into daily binned files.

---------------------------------------------------------------------------------

The Rrs spectra are available from two separate algorithms:
- Fingerprint (FP), Simis and Olsson 2013, RSE
- 3C, Groetsch et al. 2017, OE

--------------------------------------------------------------------------------

The FP and 3C outputs have different structures, quality control chains and plot functions: use the -a argument to specify which should be used.

For command line options please see
    python sorad_retrieval_qc_plot_test.py -h


For example,
    python sorad_retrieval_qc_plot_test.py -a 3c -p PML_SR004 -i "2021-08-16 00:00:00" -e "2021-08-22 23:59:59" -r -m -c -g --bbox 35 -10 40 -8
will plot and save all results from So-Rad platform PML_SR004 for the given time period, with Rrs derived from the 3C algorithm.
Files will be written to a default folder in the current directory (override this by providing an alternative path with the -t argument).
An optional bounding box is specified providing four space-separated coordinates (corner 1 lat lon corner 2 lat lon).

-------------------------------------------------------------------------------

Default quality control chain for 3C:
       (i)   Radiometric filtering (lt/ed glint filter, ed and ls anomaly filters)
       (ii)  Algorithmic filtering (3c residual and termination of rho_dd, rho_ds, or rho_ds at optimizion bounds)
       (iii) Rrs-based filtering (similarity spectrum, and filter based on range of Rrs)

Default quality control chain for FP:
       (i)   Radiometric filtering (lt/ed glint filter, ed and ls anomaly filters)
       (ii)  Rrs-based filtering (similarity spectrum, and filter based on max/min of Rrs)
       Note: `algorithmic filtering' is already applied to FP data on the geoserver (based on rho_s bounds)

Optional filters (commented out below):
       (i) Rrs shape based filtering (currently includes `coastal water filter' from Warren et al. 2019).
           Users are encouraged to apply filters for their local water type
       (ii) Variability filtering (based on z-score), following Groetsch et al. 2017.

------------------------------------------------------------------------------

Tom Jordan - tjor@pml.ac.uk - Feb 2022
"""

import sys
import os
import numpy as np
from monda.sorad import access, plots, qc
import datetime
import logging
import pandas as pd
import argparse
from math import ceil

log = logging.getLogger('sorad-test')
myFormat = '%(asctime)s | %(name)s | %(levelname)s | %(message)s'
formatter = logging.Formatter(myFormat)
logging.basicConfig(level = 'INFO', format = myFormat, stream = sys.stdout)


def run_example(platform_id = 'PML_SR004',
                start_time = datetime.datetime(2021,10,21,0,0,0),
                end_time   = datetime.datetime(2021,10,22,23,59,59),
                bbox = None,
                target='.', rrsalgorithm='fp',
                output_radiance = False,
                output_metadata = False,
                output_rrs = False,
                output_plots = False):

    """
    Download data from a specific So-Rad platform processed to Rrs using the Fingerprint (fp) or 3C (3c) algorithm.
    Apply quality control filters and plot results, save output in daily bins between requested start and end times

    platform_id: the serial number of a So-Rad platform, or None
    start_time:  date/time to start collecting data from
    end_time:    date/time to start collecting data from
    bbox:        Bounding box corner coordinates (lat,lon,lat,lon)
    target:      destination folder for plots/data
    rrsalgorithm Fingerprint (fp) or 3c Rrs processing algorithm
    """

    wl_output = np.arange(300, 1001, 1)  # output range to interpolate (ir)radiance spectra. Note that Rrs is already interpolated to 1-nm over the available sensor range.

    if rrsalgorithm == 'fp':
        layer = 'rsg:sorad_public_view_fp_full'
    elif rrsalgorithm == '3c':
        layer = 'rsg:sorad_public_view_3c_full'
    else:
        log.error(f"{rrsalgorithm} is not a valid choice and must be one of [fp, 3c]")

    initial_day = start_time.date()
    final_day = end_time.date()
    assert final_day >= initial_day

    # split the request into discrete days
    days = (final_day - initial_day).days + 1
    log.info(f"Request spans {days} day(s)")

    for i in range(days):
        this_day      = initial_day + datetime.timedelta(days = i)
        datetime_i    = datetime.datetime(this_day.year, this_day.month, this_day.day, 0, 0, 0)
        datetime_e    = datetime.datetime(this_day.year, this_day.month, this_day.day, 23, 59, 59, 999999)

        if datetime_i < start_time:
            datetime_i = start_time
        if datetime_e > end_time:
            datetime_e = end_time

        log.info(f"Request timeframe {datetime_i.isoformat()} - {datetime_e.isoformat()}")

        response = access.get_wfs(platform = platform_id,
                                  timewindow = (datetime_i, datetime_e),
                                  layer=layer, bbox=bbox)


        log.info(f"{response['length']} features received.")

        if response['length'] == 0:
            continue

        file_id = f"{datetime_i.strftime('%Y-%m-%d')}_{rrsalgorithm.upper()}"   # labelling for output data files and plots

        rrswl, time, lat, lon, rel_view_az,\
               ed, ls, lt, rrs,\
               sample_uuids, platform_ids, platform_uuids,\
               gps_speeds, tilt_avgs, tilt_stds = unpack_response(response, rrsalgorithm, wl_output)

        if output_plots:
            log.info("Creating (ir)radiance plots")
            plots.plot_ed_ls_lt(ed, ls, lt, time, wl_output, file_id, target)


        # Step (i) radiometric quality control filters (i.e. QC applied to l or e spectra)
        q_lt_ed = qc.qc_lt_ed_filter(ed, lt, time, wl_output, threshold = 0.020) # lt/Ed ratio (glint) filtering
        q_ed =    qc.qc_ed_filter(ed, min_ed_threshold = 500) # filters on ed and ls anomalies
        q_ls =    qc.qc_ls_filter(ls, wl_output, threshold = 1)
        q_rad =   qc.combined_filter(qc.combined_filter(q_lt_ed, q_ed), q_ls) # combined `radiometric' qc mask

        # Step (ii)  addtional qc metrics that apply to Rrs spectrum
        q_ss =        qc.qc_SS_NIR_filter(rrswl, rrs, upperthreshold = 3, lowerthreshold = 0.5)  # similarity spectrum
        q_maxrange =  qc.qc_rrs_maxrange(rrs, upperthreshold = 0.1, lowerthreshold = 0.00)    # filters on max and min rrs
        q_min =       qc.qc_rrs_min(rrs, rrswl)

        # (3c) algorithmic qc filters specfic to 3C (rmsd or termination at rho bounds)
        if rrsalgorithm == '3c':
            rmsd_3c =     np.array([response['result'][i]['c3_rmsd'] for i in range(len(response['result']))]) # rmsd values used in 3C residual filter
            rho_ds =      np.array([response['result'][i]['c3_rho_ds'] for i in range(len(response['result']))]) # rho factors
            rho_dd =      np.array([response['result'][i]['c3_rho_dd'] for i in range(len(response['result']))])
            rho_s =       np.array([response['result'][i]['c3_rho_s'] for i in range(len(response['result']))])
            q_rho =       qc.qc_3c_rho_filter(rho_ds, rho_dd, rho_s, upperbound = 0.1)
            q_rad_resid = qc.qc_3cresidual(q_rad, rmsd_3c, tol = 1.5)
            q_rad_3c =    qc.combined_filter(q_rho, q_rad_resid)

        if rrsalgorithm == '3c':
            q_rad_rrs = qc.combined_filter(q_rad_3c, qc.combined_filter(q_min, (qc.combined_filter(q_ss, q_maxrange)))) # recommended rrs qc mask for 3C method (combines step (i), (ii) and (iii) QC)
        elif rrsalgorithm == 'fp':
            q_rad_rrs = qc.combined_filter(q_rad,    qc.combined_filter(q_min, (qc.combined_filter(q_ss, q_maxrange)))) # recommended rrs qc mask for FP method (combines step (i) and (ii) QC)

        # Optional filters
        # q_coastal = qc_coastalwater_rrsfilter(rrs, wl) #  filter based on expected shape of rrs - example from Warren 2019 used. users can input their own spectra here (will depend on water type)
        # q_var = qc_radiometric_variability(ed, lt, ls, time, wl, windowlength = 60, var_threshold =1.1, var_metric = 'zscore_max')

        if output_plots and rrsalgorithm == 'fp':
            log.info("Creating Rrs plots")
            plots.plot_rrs_qc_fp(rrs, time, rrswl, q_rad, q_rad_rrs,              file_id, target)
            plots.plot_coveragemap(lat, lon, q_rad, file_id, target, map_resolution = 11)
            plots.plot_results(ed, ls, wl_output, rrs, rrswl, time, q_rad_rrs, file_id, target)

        elif output_plots and rrsalgorithm == '3c':
            log.info("Creating Rrs plots")
            plots.plot_rrs_qc_3c(rrs, time, rrswl, q_rad, q_rad_3c, q_rad_rrs, file_id, target)
            plots.plot_coveragemap(lat, lon, q_rad_3c, file_id, target, map_resolution = 11)
            plots.plot_results(ed, ls, wl_output, rrs, rrswl, time, q_rad_rrs, file_id, target)


        d = pd.DataFrame()   # store core metadata and qc flags in a data frame for easy output formatting
        d['sample_uuid'] = sample_uuids
        d['platform_id'] = platform_ids
        d['platform_uuid'] = platform_uuids
        d['timestamp'] = time
        d['lat'] = lat
        d['lon'] = lon
        d['gps_speed'] = gps_speeds
        d['tilt_avg'] = tilt_avgs
        d['tilt_std'] = tilt_stds
        d['rel_view_az '] = rel_view_az
        d['q_1'] = q_rad     # Mask after step (i) QC

        if rrsalgorithm == 'fp':
            d['q_2'] = q_rad_rrs # Mask after step (ii) QC. Currently recommended for FP rrs data analysis

        elif rrsalgorithm == '3c':
            d['q_2'] = q_rad_3c   # Mask after step (ii) QC
            d['q_3'] = q_rad_rrs  # Mask after step (iii) QC: currently recommended for 3C  rrs data analysis

        # optional to output all qc masks
        # q_keys = [i for i in locals() if i.startswith('q_')]
        # for i in range(len(q_keys)): # add all qc fields to the dataframe
        #     d[str(q_keys[i])] = pd.Series(eval(q_keys[i]))

        # Store outputs
        if output_metadata:
            d_filename = os.path.join(target, file_id + '_metadata.csv')
            if os.path.exists(d_filename):
                log.warning(f"File {d_filename} was overwritten")
            d.to_csv(d_filename)

        if output_rrs:
            r_filename = os.path.join(target, file_id + '_Rrs.csv')
            if os.path.exists(r_filename):
                log.warning(f"File {r_filename} was overwritten")
            np.savetxt(r_filename, rrs, delimiter=',',  header = ",".join([str(w) for w in rrswl]), fmt='%.8f')

        if output_radiance:
            header = ",".join([str(w) for w in wl_output])
            ls_filename = os.path.join(target, file_id + '_Ls.csv')
            lt_filename = os.path.join(target, file_id + '_Lt.csv')
            ed_filename = os.path.join(target, file_id + '_Ed.csv')
            if os.path.exists(ls_filename):
                log.warning(f"File {ls_filename} was overwritten")
            if os.path.exists(lt_filename):
                log.warning(f"File {lt_filename} was overwritten")
            if os.path.exists(ed_filename):
                log.warning(f"File {ed_filename} was overwritten")
            np.savetxt(ls_filename, ls, delimiter=',', header = header, fmt='%.8f')
            np.savetxt(lt_filename, lt, delimiter=',', header = header, fmt='%.8f')
            np.savetxt(ed_filename, ed, delimiter=',', header = header, fmt='%.8f')

    return response


def unpack_response(response, rrsalgorithm, wl_out):
    """
    Unpack the WFS response
    """
    #log.info(response['result'][0].keys())   # uncomment to show all available fields

    time          = [response['result'][i]['time'] for i in range(len(response['result']))]
    lat           = np.array([response['result'][i]['lat'] for i in range(len(response['result']))])
    lon           = np.array([response['result'][i]['lon'] for i in range(len(response['result']))])
    rel_view_az   = np.array([response['result'][i]['rel_view_az'] for i in range(len(response['result']))])
    sample_uuid   = np.array([response['result'][i]['sample_uuid'] for i in range(len(response['result']))])
    platform_id   = np.array([response['result'][i]['platform_id'] for i in range(len(response['result']))])
    platform_uuid = np.array([response['result'][i]['platform_uuid'] for i in range(len(response['result']))])
    gps_speed     = np.array([response['result'][i]['id'] for i in range(len(response['result']))])
    tilt_avg      = np.array([response['result'][i]['tilt_avg'] for i in range(len(response['result']))])
    tilt_std      = np.array([response['result'][i]['tilt_std'] for i in range(len(response['result']))])

    ed = access.get_l1spectra(response, 'ed_', wl_out) # # irradiance spectra in 2D matrix format: rows time index, columns wavelength
    ls = access.get_l1spectra(response, 'ls_', wl_out)
    lt = access.get_l1spectra(response, 'lt_', wl_out)


    if rrsalgorithm == '3c':
        rrswl = np.arange(response['result'][0]['c3_wl_grid'][0], response['result'][0]['c3_wl_grid'][1], response['result'][0]['c3_wl_grid'][2])  # reconstruct wavelength grid for Rrs
        rrs = np.array([response['result'][i]['c3_rrs'][:] for i in range(len(response['result']))]) # 2D matrix format: rows time index, columns wavelength

    elif rrsalgorithm == 'fp':
        rrswl  = np.arange(response['result'][0]['wl_grid'][0], response['result'][0]['wl_grid'][1]-1, response['result'][0]['wl_grid'][2])  # reconstruct wavelength grid for Rrs
        rrs_    = np.array([response['result'][i]['rrs'][:] for i in range(len(response['result']))])  # rrs spectra 2D matrix format: rows time index, columns wavelength
        offset = np.array([response['result'][i]['nir_offset'] for i in range(len(response['result']))])
        rrs = np.array([rrs_[i,:] - np.ones(len(rrswl))*offset[i] for i in range(len(rrs_))]) # spectral offset (applied as default definition of FP rrs)

    return rrswl, time, lat, lon, rel_view_az, ed, ls, lt, rrs, sample_uuid, platform_id, platform_uuid, gps_speed, tilt_avg, tilt_std


def parse_args():
    """Interpret command line arguments"""
    parser = argparse.ArgumentParser()
    parser.add_argument('-a','--algorithm',   required = False, type=str, default = 'fp', help = "rrs processing algorithm: fp or 3c")
    parser.add_argument('-p','--platform',    required = False, type = str, default = 'PML_SR004', help = "Platform serial number, e.g. PML_SR004.")
    parser.add_argument('-i','--start_time',  required = False, type = lambda s: datetime.datetime.strptime(s, '%Y-%m-%d %H:%M:%S'),
                                              default = datetime.datetime(2021,10,21,0,0,0),
                                              help = "Initial UTC date/time in format 'YYYY-mm-dd HH:MM:SS'")
    parser.add_argument('-e','--end_time',    required = False, type = lambda s: datetime.datetime.strptime(s, '%Y-%m-%d %H:%M:%S'),
                                              default = datetime.datetime(2021,10,22,23,59,59),
                                              help = "Final UTC date/time in format 'YYYY-mm-dd HH:MM:SS'")
    parser.add_argument('-b','--bbox',        required = False, type = float, nargs='+', default = None, help = "Restrict query to bounding box format [corner1lat corner1lon corner2lat corner2lon]")
    parser.add_argument('-t','--target',      required = False, type = str, default=None,
                                              help = "Path to target folder for plots (defaults to 'So-Rad_testoutput' in the current folder).")
    parser.add_argument('-r','--output_radiance',  required = False, action='store_true', help = "Output Ls, Lt, Ed spectra to csv file")
    parser.add_argument('-m','--output_metadata',  required = False, action='store_true', help = "Output metadata to csv file")
    parser.add_argument('-c','--output_rrs',  required = False, action='store_true', help = "Output Rrs spectra to csv file")
    parser.add_argument('-g','--output_plots',  required = False, action='store_true', help = "Output plots")

    args = parser.parse_args()

    return args


if __name__ == '__main__':

    args = parse_args()

    if not any([args.output_radiance, args.output_metadata, args.output_rrs, args.output_plots]):
        log.warning("No plots or data outputs specified (see -h for help)")

    if args.target is None:
        args.target = os.path.join('.', 'So-Rad_test-output')

    if not os.path.isdir(args.target):
        os.mkdir(args.target)

    response = run_example(args.platform, args.start_time, args.end_time, args.bbox, args.target, args.algorithm.lower(),
                           args.output_radiance, args.output_metadata, args.output_rrs, args.output_plots)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""

Quality control functions for monda So-Rad data download

Quality control functions output a logical mask using convention:
 1 == valid timestamp, 0 == not valid timestamp.

------------------------------------------------------------------------------

Tom Jordan - tjor@pml.ac.uk - Feb 2022
"""

import numpy as np
import datetime
import pandas as pd


# sub routines
def wl_find(wl, target):
    "Finds wavelength index nearest to target"
    return np.argmin((np.array(wl)-target)**2)


def combined_filter(q_1, q_2):
   """Combines 2 QC masks - tests for failures subject to logical OR condition"""

   q_combined = np.ones(len(q_1))
   for i in range(len(q_1)):
      if q_1[i] == 0 or q_2[i] == 0:
         q_combined[i] = 0

   return q_combined


def nearest(items, pivot):
     """Sub function used to locate nearest values indicies in list x relative to a pivot - used within rolling variability """

     nearest_value = min(items, key=lambda x: abs(x - pivot)) # finds nearest value
     nearest_index = items.index(min(items, key=lambda x: abs(x - pivot))) # finds nearest index

     return nearest_value, nearest_index


# Filters used in QC of l and e spectra (Step 1 in 3C and FP chain)
def qc_lt_ed_filter(ed, lt, time, wl, threshold = 0.020):
    """Funtion to filter by lt_ed ratio in NIR: basic implementation using absolute threshold defined in sr^-1 on [850, 950] nm"""

    lt_ed = np.divide(lt, ed)
    lambda_l = wl_find(800, wl)
    lambda_h = wl_find(950, wl)
    lw_ratio = np.nanmax(lt_ed[:,lambda_l:lambda_h], axis=1)

    q_ratio =  np.ones(len(ed))
    q_ratio[lw_ratio > threshold] = 0

    return q_ratio


def qc_ed_filter(ed, min_ed_threshold = 500):
    """Function to filter by the minimum (spectral max) ed value"""

    q_ed =  np.ones(len(ed)) # qc
    spec_max_ed = np.nanmax(ed.T, 0) # spectral max for Ed
    q_ed[spec_max_ed < min_ed_threshold] = 0

    return q_ed


def qc_ls_filter(ls, wl, threshold = 1):
    """Funtion to filter out ls spectra that are anomalously high in NIR"""

    lambda_1l = wl_find(550, wl) # central wavelength band
    lambda_1h = wl_find(580, wl)
    lambda_2l = wl_find(800, wl) # long wavelength band
    lambda_2h = wl_find(950, wl)

    ls_ratio = np.nanmax(ls[:,lambda_2l:lambda_2h], axis=1) / np.nanmax(ls[:,lambda_1l:lambda_1h], axis=1)

    q_ls =  np.ones(len(ls)) # qc mask: 1=valid, 0=not valid
    q_ls[ls_ratio > threshold] = 0

    return q_ls


# 3C-specific filtering (step 2 in 3C QC chain)
def qc_3cresidual(q_rad,resid,tol = 3):
    """3C residual fliter based on standard deviation of daily rmsd ditribution that has
       already passed radiometric quality control. tol is a standard deviation multiple
       with 3*sigma set to default"""

    q_rad_3c = q_rad
    if np.sum(q_rad) > 0:
        resid_max = np.ones(len(q_rad))*(tol*np.nanstd(resid[q_rad==1]) + np.nanmean(resid[q_rad==1]))
        q_rad_3c[resid > resid_max] = 0

    return q_rad_3c


def qc_3c_rho_filter(rho_ds, rho_dd, rho_s, upperbound = 0.1):
    """Function to remove 3C retreivals where rho factors terminate on upper optimzation bound
    (note: not strictly required for lower optimzation bound)"""

    q_rho3c =  np.ones(len(rho_ds))

    q_rho3c[np.abs(rho_ds  - upperbound) < 0.001] = 0
    q_rho3c[np.abs(rho_dd  - upperbound) < 0.001] = 0
    q_rho3c[np.abs(rho_s - upperbound) < 0.001] = 0

    return q_rho3c


# filters used in qc of Rrs spectra: Step 3 in 3C chain and 2 in FP chain)
def qc_SS_NIR_filter(wl, rrs, upperthreshold = 3, lowerthreshold = 0.5):
    """Filter based on NIR similarity spectrum (reflectance ratio of 779 and 865 nm) in Ruddick et al. 2006."""

    lambda779 = wl_find(wl, 779)
    lambda865 = wl_find(wl, 865)

    r779_865 = rrs[:,lambda779]/rrs[:,lambda865]

    q_ss =  np.ones(len(rrs)) # qc mask: 1=valid, 0=not valid
    q_ss[r779_865  > upperthreshold] = 0
    q_ss[r779_865  < lowerthreshold] = 0

    return q_ss


def qc_rrs_maxrange(rrs, upperthreshold = 0.1, lowerthreshold = 0.0):
    """ Filter based on range of rrs maximum """

    q_maxrange = np.ones(len(rrs))

    rrs_max = np.nanmax(rrs,axis=1)
    q_maxrange[rrs_max > upperthreshold] = 0
    q_maxrange[rrs_max < lowerthreshold] = 0

    return q_maxrange


def qc_rrs_min(rrs, wl):
    """ Replicates filtering step in FP processing that Rrs must be > 0 on [370,700] nm
    but applied to Rrs with offset included"""

    q_min = np.ones(len(rrs))

    lambda375 = wl_find(wl,375)
    lambda700= wl_find(wl,700)

    rrs_min = np.nanmin(rrs[:,lambda375:lambda700],axis=1)
    q_min[rrs_min <  0] = 0

    return q_min


# optional filters (not currently applied in FP and 3C chain)
def rolling_variability(spectrum, timestamp ,window_time, wl_range):
    """ Sub funtion to calculate rolling spectral variability metrics based on z-score (std normalized spectra) and percentage differences (mean normalized spectra).
    Based on function in Groetsch et. al 2017

    # Work flow:
    # (i) converts timestamp and spectra to data frame format (indexed by timestamp, rows wavelengths)
    # (ii) calculates rolling mean and standrad deviations using dataframe method
    # (iii) re-references rolling mean/std to window center
    # (iv) computes rolling z-score and percentage difference metrics
    # Input - window size in minutes and l or e spectrum
    # Notes - data gaps in rolling mean/std are handled via linear interpolation (default setting of rolling functions in pandas)
    """

    # (i) dataframe format
    window_time_secs = window_time*60 # converts to secs
    timeindex = [pd.Timestamp(timestamp[i]) for i in range(len(timestamp))] # converts to pandas format
    spectrum_DF = pd.DataFrame(spectrum,timeindex,columns = wl_range) # creates dataframe for spectum indexed by timestamp

    # (ii) Rolling means
    spectrum_Rmean = np.array(spectrum_DF.rolling(window=str(window_time)+'min').mean()) # rolling mean of spectrum
    spectrum_Rstd = np.array(spectrum_DF.rolling(window=str(window_time)+'min').std()) # rolling std of spectrum

    # (iii) Reference rolling mean and std to center of moving window (this is required as window center CANNOT be applied to ragged datetime index)
    timestamp_center = [timestamp[i]+datetime.timedelta(seconds=round(window_time_secs/2)) for i in range(len(timestamp))] # time-stamp shifted to window center
    spectrum_Rmean_centered = np.nan*np.ones([len(timestamp),len(wl_range)]) # centered versions of rolling mean and std
    spectrum_Rstd_centered = np.nan*np.ones([len(timestamp),len(wl_range)])
    for i in range(len(timestamp)):
        centertime = nearest(timestamp,timestamp_center[i]) # finds time and index of window center using nearest function
        centerindex = int(centertime[1]) # converts nearest index to int
        spectrum_Rmean_centered[i,:] = spectrum_Rmean[centerindex,:]
        spectrum_Rstd_centered[i,:] = spectrum_Rstd[centerindex,:]

    # (iv) Caculate moving z-score and percentage-differences
    pct_diff = np.nan*np.ones([len(timestamp),len(wl_range)]) # matrix for rolling percentage difference
    pct_diff_max = np.nan*np.ones(len(timestamp)) # spectral max of rolling pct difference
    pct_diff_mean = np.nan*np.ones(len(timestamp)) # spectral mean of rolling pct difference
    zscore = np.nan*np.ones([len(timestamp),len(wl_range)]) # matrix of (absolute) rolling z-score
    zscore_max =np.nan*np.ones(len(timestamp)) # spectral max of rolling z-score
    zscore_mean = np.nan*np.ones(len(timestamp)) # spectral mean of rolling z-score
    for i in range(len(timestamp)):
        pct_diff[i,:] = np.abs(100*(spectrum[i,:]-spectrum_Rmean_centered[i,:])/spectrum_Rmean_centered[i,:]) # rolling percentage difference
        pct_diff_max[i] = np.nanmax(pct_diff[i,:]) # spectral max
        pct_diff_mean[i] = np.nanmean(pct_diff[i,:]) # spectral mean
        zscore[i,:] = np.abs((spectrum[i,:]-spectrum_Rmean_centered[i,:])/spectrum_Rstd_centered[i,:]) # rollling absolute z-score
        zscore_max[i] = np.nanmax(zscore[i,:]) # spectral max
        zscore_mean[i] = np.nanmean(zscore[i,:]) # spectral mean

    return pct_diff_max, pct_diff_mean, zscore_max, zscore_mean


def qc_radiometric_variability(ed, lt, ls, time, wl, windowlength = 60, var_threshold = 1.1, var_metric = 'zscore_max'):
   """Filters out local variability anomalie of input l and e spectra based on rolling
      variability metric (spectral maxium of absolute z-score as default) """

   if var_metric ==  'pct_diff_max':
        index = 0
   elif var_metric ==  'pct_diff_mean':
        index = 1
   elif var_metric ==  'zscore_max':
       index =  2
   elif var_metric ==  'zscore_mean':
       index =  3

   ls_var = rolling_variability(ls, time, windowlength, wl)[index]  # caculates rolling z-score (spectral maximum)
   lt_var = rolling_variability(lt, time, windowlength, wl)[index]  # can also select pct_diff_max pct_diff_mean, zscore_max, zscore_mean using
   ed_var = rolling_variability(ed, time, windowlength, wl)[index]  # [0], [1], [3] indicies

   var_max = np.array([np.nanmax([ls_var[i], lt_var[i], ed_var[i]])  for i in range(len(ed))])  # takes `worst case variability metric' of 3 sensors

   q_var =  np.ones(len(ed))  # qc mask: 1=valid, 0=not valid
   q_var[var_max > var_threshold] = 0

   return q_var


def qc_coastalwater_rrsfilter(rrs, wl):
    """Rrs-shape based filter from Warren et al. 2019"""

    q_coastal =  np.ones(len(rrs)) # qc mask: 1=valid, 0=not valid

    lambda_350 = wl_find(350,wl)   # filter 1:  remove mean rrs on [350, 400] nm  < -0.0005 sr^-1
    lambda_400 = wl_find(400,wl)
    mean_rrs_350_400 = np.nanmean(rrs[:,lambda_350:lambda_400],axis=1)
    q_coastal[mean_rrs_350_400 < -0.0005] = 0

    lambda_800 = wl_find(800,wl)   # filter 2: remove mean rrs on [800, 900] nm < -0.0005 sr^-1
    lambda_900 = wl_find(900,wl)
    mean_rrs_800_900 = np.nanmean(rrs[:,lambda_800:lambda_900],axis=1)
    q_coastal[mean_rrs_800_900 < -0.0005] = 0

    max_rrs = np.max(rrs,axis=1)   # filter 3: remove max rrs > 0.015 sr^-1
    q_coastal[max_rrs > 0.015] = 0

    lambda_760 = wl_find(760,wl) # filter 4: oxygen band SNR filter - remove
    lambda_770 = wl_find(770,wl) # local peak intensities on [760, 770] nm that
    lambda_560 = wl_find(560,wl) # are > 10 %  of the green rrs peak
    lambda_600 = wl_find(600,wl)
    NIR_peak = np.max(rrs[:,lambda_760:lambda_770],axis=1) - np.min(rrs[:,lambda_760:lambda_770],axis=1)
    green_peak =  np.max(rrs[:,lambda_560:lambda_600],axis=1)
    q_coastal[NIR_peak/green_peak  > lambda_600] = 0

    max_rrs_index = np.argmax(rrs,axis=1) # filter 5: remove case when peak rrs occurs > 600 nm
    lambda_600 = wl_find(600,wl)
    q_coastal[max_rrs_index > lambda_600] = 0

    return q_coastal



import setuptools
from setuptools import setup

with open("README.md", "r") as long_info:
    long_description = long_info.read()

setup(
    name='monda',
    version='0.2.0',
    author='',
    author_email='',
    description='A package for retrieval, quality control and analysis of Data from MONOCLE systems',
    url='https://github.com/monocle-h2020/MONDA',
    py_modules=["monda"],
    package_dir={'':'src'},
    packages=setuptools.find_packages(where="src"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=["Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",],
    python_requires=">=3.8",
    install_requires = ['markdown','requests','cartopy','matplotlib','numpy', 'pandas', 'datetime',
                        'argparse'],
)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""

Plot functions for monda So-Rad data download illustrating: (ir)radiance spectra, quality
control procedures applied to the processing chain, Rrs results, atmospheric conditions
(Ls/Ed(400) ratio) and coverage map.

------------------------------------------------------------------------------
Tom Jordan - tjor@pml.ac.uk - Feb 2022.
"""


import os
import sys
import logging
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import matplotlib.cm as cm

import cartopy.crs as ccrs
import cartopy.io.img_tiles as cimgt
from cartopy.mpl.ticker import LongitudeFormatter, LatitudeFormatter
from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER

log = logging.getLogger('sorad-plotter')
myFormat = '%(asctime)s | %(name)s | %(levelname)s | %(message)s'
formatter = logging.Formatter(myFormat)
logging.basicConfig(level = 'INFO', format = myFormat, stream = sys.stdout)


# sub routines
def wl_find(wl, target):
    "Finds wavelength index nearest to target"
    return np.argmin((np.array(wl)-target)**2)

def percentile_amplitude(values):
    "determine the 99th percentile of a range to limit y axis range limit"
    values =values[np.isfinite(values)]
    values = values[values>=0.0]
    return np.percentile(values, 99)

def outlier_limit(values):
    "determine outlier amplitude to limit y axis range limit"
    values = values[np.isfinite(values)]
    values = values[values>=0.0]
    outlier_limit = np.median(values) + 1.5 * np.std(values)
    ylimit = np.max(values[values<outlier_limit])
    return ylimit

def find_max_in_wl_range(wl, spectra, minwl, maxwl):
    "find peak amplitude wihtin a wavelength range for a set of spectra"
    wlmin = wl_find(wl, 400)
    wlmax = wl_find(wl, 800)
    maxima = np.amax(spectra[:, wlmin:wlmax], axis=0)
    return maxima

# plot functions
def plot_ed_ls_lt(ed, ls, lt, time, wl, file_id, target):
    """Spectral plots for ed ls lt"""

    plt.figure(figsize=(8,16))
    plt.rc('font', size=16)
    plt.suptitle(str(file_id))

    plt.subplot(3,1,1)
    plt.plot(wl,ed.T,linewidth=0.4,alpha=0.6)
    plt.xlim(350,900)
    plt.gca().set_ylim(bottom =-0.001)
    ymax = outlier_limit(find_max_in_wl_range(wl, ed, 400, 800))  # limit y axis based on maximum values < outlier range
    plt.gca().set_ylim(top = ymax)
    #plt.grid()
    plt.xlabel('Wavelength [nm]')
    plt.ylabel('$E_{d}$ [mW m$^{-2}$ nm$^{-1}$]')

    plt.subplot(3,1,2)
    plt.plot(wl,ls.T,linewidth=0.4,alpha=0.6)
    plt.xlim(350,900)
    plt.gca().set_ylim(bottom =-0.001)
    ymax = outlier_limit(find_max_in_wl_range(wl, ls, 400, 800))  # limit y axis based on maximum values < outlier range
    plt.gca().set_ylim(top = ymax)
    #plt.grid()
    plt.xlabel('Wavelength [nm]')
    plt.ylabel('$L_{s}$ [mW m$^{-2}$ sr$^{-1}$ nm$^{-1}$]')

    plt.subplot(3,1,3)
    plt.plot(wl,lt.T,linewidth=0.4,alpha=0.6)
    plt.xlim(350,900)
    plt.gca().set_ylim(bottom =-0.001)
    ymax = outlier_limit(find_max_in_wl_range(wl, lt, 400, 800))  # limit y axis based on maximum values < outlier range
    plt.gca().set_ylim(top = ymax)
    #plt.grid()
    plt.xlabel('Wavelength [nm]')
    plt.ylabel('$L_{t}$ [mW m$^{-2}$ sr$^{-1}$ nm$^{-1}$]')

    plt.tight_layout()
    plt.savefig(os.path.join(target, file_id + '_Ed-Ls-Lt-spectra.png'), format='png', dpi=150)

    return


def plot_rrs_qc_3c(rrs, time, wl, q_1, q_2, q_3, file_id, target):
    """ rrs plot function showing sequential quality control filters"""

    if np.sum(q_3) > 0:
        plt.figure()
        plt.figure(figsize=(12,8))
        plt.suptitle(str(file_id))

        ymax = np.ceil(np.nanmax(rrs.T[:,q_2==1]*1000))/1000;

        plt.subplot(2, 2, 1)
        plt.title(f"No QC (n = {int(len(q_1))})")
        plt.plot(wl, rrs.T, linewidth=0.4, alpha=0.6)
        plt.xlim(350, 950)
        plt.ylim(-0.002, ymax) # force axis limits
        #plt.grid()
        plt.xlabel('Wavelength [nm]')
        plt.ylabel('$R_{rs}$ [sr$^{-1}$]')

        plt.subplot(2, 2, 2)
        plt.title(f"Rad. QC (n = {int(np.sum(q_1))})")
        plt.plot(wl, rrs.T[:,q_1==1], linewidth=0.4, alpha=0.8)
        plt.xlim(350, 950)
        plt.ylim(-0.002, ymax) # force axis limits
        #plt.grid()
        plt.xlabel('Wavelength [nm]')

        plt.subplot(2, 2, 3)
        plt.title(f"Rad. + 3C QC (n = {int(np.sum(q_2))})")
        plt.plot(wl,rrs.T[:,q_2==1], linewidth=0.4, alpha=0.8)
        plt.xlim(350, 950)
        plt.ylim(-0.002, ymax) # force axis limits
        #plt.grid()
        plt.xlabel('Wavelength [nm]')
        plt.ylabel('$R_{rs}$   [sr$^{-1}$]')

        plt.subplot(2, 2, 4)
        plt.title(f"Rad. + 3C + Rrs QC (n = {int(np.sum(q_3))})")
        plt.plot(wl,rrs.T[:,q_3==1], linewidth=0.4, alpha=0.8)
        plt.xlim(350, 950)
        plt.ylim(-0.002, ymax) # force axis limits
        #plt.grid()
        plt.xlabel('Wavelength [nm]')

        plt.subplots_adjust(hspace=0.5)

        plt.savefig(os.path.join(target, file_id + '_Rrs_QC.png'), format='png', dpi=150)

    return


def plot_rrs_qc_fp(rrs, time, wl, q_1, q_2, file_id, target):
    """ rrs plot function showing sequential quality control filters"""

    if np.sum(q_2) > 0:
        plt.figure()
        plt.figure(figsize=(12, 8))
        plt.suptitle(str(file_id))

        ymax = np.ceil(np.nanmax(rrs.T[:,q_2==1]*1000))/1000;

        plt.subplot(2, 2, 1)
        plt.title(f"No QC (n = {int(len(q_1))})")
        plt.plot(wl, rrs.T, linewidth=0.4, alpha=0.6)
        plt.xlim(350, 950)
        plt.ylim(-0.002, ymax) # force axis limits
        #plt.grid()
        plt.xlabel('Wavelength [nm]')
        plt.ylabel('$R_{rs}$ [sr$^{-1}$]')

        plt.subplot(2, 2, 2)
        plt.title(f"Rad QC (n = {int(np.sum(q_1))})")
        plt.plot(wl, rrs.T[:,q_1==1], linewidth=0.4, alpha=0.8)
        plt.xlim(350, 950)
        plt.ylim(-0.002, ymax) # force axis limits
        #plt.grid()
        plt.xlabel('Wavelength [nm]')

        plt.subplot(2, 2, 3)
        plt.title(f"Rad + Rrs QC (n = {int(np.sum(q_2))})")
        plt.plot(wl, rrs.T[:,q_2==1], linewidth=0.4, alpha=0.8)
        plt.xlim(350, 950)
        plt.ylim(-0.002, ymax) # force axis limits
        #plt.grid()
        plt.xlabel('Wavelength [nm]')

        plt.subplots_adjust(hspace=0.5)

        plt.savefig(os.path.join(target, file_id + '_Rrs_QC.png'), format='png', dpi=150)

    return


def plot_coveragemap(lat, lon , q, file_id, target, map_resolution=11):
    """ coverage map showing quality control filtered data: color scheme matches
    `results' plot"""
    if np.sum(q) > 0:
        colors= cm.cool(np.linspace(0, 1, int(sum(q))))

        plt.figure(figsize=(15,10))
        extent = [np.floor(np.min(lon*100))/100, np.ceil(np.max(lon*100))/100, np.floor(np.min(lat*100))/100, np.ceil(np.max(lat*100))/100]
        request = cimgt.StamenTerrain()
        ax = plt.axes(projection=ccrs.PlateCarree())
        ax.add_image(request, map_resolution)
        ax.set_extent(extent, ccrs.PlateCarree())
        gl = ax.gridlines(draw_labels=True)
        gl.top_labels_top = gl.right_labels = False
        gl.xformatter =  LONGITUDE_FORMATTER
        gl.yformatter =  LATITUDE_FORMATTER
        gl.xlabel_style = {'size': 12,  'rotation':45}
        gl.ylabel_style = {'size': 12,  'rotation': 0}

        lon_formatter = LongitudeFormatter(zero_direction_label=True)
        lat_formatter = LatitudeFormatter()
        ax.xaxis.set_major_formatter(lon_formatter)
        ax.tick_params(labelsize=10)
        plt.scatter(lon,lat,s=8,color='gray',transform=ccrs.PlateCarree(), label='Failed QC')
        plt.scatter(lon[q==1][0],lat[q==1][0],s=15,color=colors[0],transform=ccrs.PlateCarree(),label='Passed QC')
        for i in range(int(sum(q))):
            plt.scatter(lon[q==1][i],lat[q==1][i],s=15,color=colors[i],transform=ccrs.PlateCarree())

        plt.rc('font', size=14)
        plt.title(str(file_id))
        plt.legend()

        plt.savefig(os.path.join(target, file_id + '_coverage-map.png'), format='png', dpi=150)

    return


def plot_results(ed ,ls, wl_out, rrs, rrswl, time, q, file_id, target):
    """ Results plot showing: (i) sky measurement conditions using ls(400)/ed(400) ratio,
    (ii) qc-filtered rrs. Panel (i) is used to illustrate timestamps passing qc. Color scale matches
    between panels so time series and rrs can be visually referenced """

    if np.sum(q) > 0:
        lambda_400 = wl_find(wl_out ,400)  # atmopsheric conditions, clear (<.25) to overcast (>.8)
        ls_ed_400 = ls[:,lambda_400] / ed[:,lambda_400]

        colors = cm.cool(np.linspace(0, 1, int(sum(q))))  # colour map to match rrs with time series
        timestamp = np.array(time)  # convert to np array for masking

        plt.figure(figsize=(10, 14))
        plt.rc('font', size=19)

        plt.subplot(2, 1, 1)
        plt.suptitle(str(file_id))

        plt.title(f"Sky conditions: $\pi Ls(400)/Ed$(400) = {np.round(np.mean(np.pi*ls_ed_400),3)} +/- {np.round(np.std(np.pi*ls_ed_400), 3)}")
        plt.gca().xaxis.set_major_formatter(mdates.DateFormatter('%H'))
        plt.ylabel('Degree of cloudiness: \n $\pi L_{s}(400)/E_{d}$(400)')
        plt.plot_date(timestamp, np.pi*ls_ed_400, color='gray', ms=1, label = 'Failed QC')
        plt.plot_date(timestamp[q==1][0], np.pi*ls_ed_400[q==1][0], color=colors[0,:], ms=3 ,label = 'Passed QC')

        for i in range(int(sum(q))):
            plt.plot_date(timestamp[q==1][i], np.pi*ls_ed_400[q==1][i], color=colors[i,:], ms=3)

        plt.ylim(0, 1.6)
        plt.legend()
        #plt.grid()
        plt.xlabel('UTC time [hrs]')

        plt.subplot(2,1,2)
        plt.title('Remote-sensing reflectance: $R_{rs}$')

        for i in range(int(sum(q))):
            plt.plot(rrswl, rrs[q==1][i,:], color=colors[i,:], linewidth=0.6, alpha=0.6)

        plt.plot(rrswl, np.nanmean(rrs[q==1], axis=0), color='black', linewidth=2, alpha=1, label='Mean spectrum')
        plt.xlim(350, 900)
        plt.gca().set_ylim(bottom =-0.001)
        #plt.grid()
        plt.xlabel('Wavelength [nm]')
        plt.ylabel('$R_{rs}$  [sr$^{-1}$]')
        plt.legend()

        plt.savefig(os.path.join(target, file_id + '_results.png'), format='png', dpi=150)

    return



import requests
import sys
import io
import csv
import logging

log = logging.getLogger('WISP-access')
myFormat = '%(asctime)s | %(name)s | %(levelname)s | %(message)s'
formatter = logging.Formatter(myFormat)
logging.basicConfig(level = 'INFO', format = myFormat, stream = sys.stdout)


def WISP_data_API_call(request_string, output_file=None):
    """ parse the request string to the cuurent WISP data service and optionally save output to file.
    See https://github.com/monocle-h2020/Access_to_WISPstation_API/blob/main/WISPcloud_API_documentation.ipynb for help on building request strings
    examples:
    WISP_data_API_call('REQUEST=GetDocumentation')
    WISP_data_API_call('REQUEST=ShowRecent')
    WISP_data_API_call('REQUEST=GetData&TIME=2018-05-11T09:00,2018-05-11T19:00&INSTRUMENT=WISPstation001')
    """
    username = 'public_access'
    password = 'WISPstation'
    Data_service = 'https://wispcloud.waterinsight.nl/api/query?SERVICE=Data&VERSION=1.0'
    url_string = f'{Data_service}&{request_string}'
    log.info(f"Request URL = {url_string}")

    if output_file is not None:
        datastring = io.StringIO(requests.get(url_string, auth=(username, password)).content.decode('utf-8'))
        print(datastring.read(), file=open(output_file, 'w'))

    datastring = io.StringIO(requests.get(url_string, auth=(username, password)).content.decode('utf-8'))
    data = list(csv.DictReader(filter(lambda row: row[0] != '#', datastring), delimiter='\t'))

    return data



import matplotlib.pyplot as plt
from matplotlib import style
import numpy as np

from monda.WISP import access

import sys
import logging

log = logging.getLogger('WISP-access')
myFormat = '%(asctime)s | %(name)s | %(levelname)s | %(message)s'
formatter = logging.Formatter(myFormat)
logging.basicConfig(level = 'INFO', format = myFormat, stream = sys.stdout)



def list_to_array(lstring):
   try:
        arr = np.array(lstring.lstrip('[').rstrip(']').split(',')).astype(np.float64)
        return arr
   except:
        return None

def plot_reflectance_station(instrument, day, start, stop, output_rrs):
    ''' Plot reflectance spectra for one WISPstation for one day
        Retrieve data from WISPcloud API
    '''
    style.use('seaborn-whitegrid')
    REQUEST = 'REQUEST=GetData&INSTRUMENT={}&include=measurement.date,measurement.id,level2.reflectance,site.name,level2.quality&TIME={}T{},{}T{}'.format(
        instrument, day, start, day, stop)

    l2r = access.WISP_data_API_call(REQUEST, output_rrs)
    rows = len(l2r)
    log.info(f"{rows} records retrieved")

    if rows == 0:
        return None

    color = iter(plt.cm.viridis(np.linspace(0, 1, rows)))
    wl = np.array(list(range(350, 901, 1)))
    fig = plt.figure(figsize=(8, 6))
    for meas in l2r:
        #print(meas.keys())
        Rrs = list_to_array(meas['level2.reflectance'])
        if Rrs is not None:
            c = next(color)
            plt.plot(wl, Rrs, c=c, label=meas['measurement.date'][10:16])
            station = meas['site.name']
    plt.xlabel('Wavelength [nm]')
    plt.ylabel('Rrs [sr-1]')
    plt.xlim(wl[0], wl[-1])
    plt.legend(loc='upper left', prop={'size': 8}, bbox_to_anchor=(1, 1))
    plt.title('{} at {} on {}'.format(instrument, station, day))
    plt.tight_layout(pad=7)

    return fig


def plot_radiances_station(instrument, day, start, stop, output_rad):
    ''' Plot (ir)radiance spectra for one WISPstation for one day
        Retrieve data from WISPcloud API
    '''
    style.use('seaborn-whitegrid')
    REQUEST = 'REQUEST=GetData&INSTRUMENT={}&include=measurement.date,measurement.id,site.name,ed.irradiance,ld.radiance,lu.radiance,level2.quality&TIME={}T{},{}T{}'.format(
        instrument, day, start, day, stop)

    l1b = access.WISP_data_API_call(REQUEST, output_rad)
    rows = len(l1b)
    log.info(f"{rows} records retrieved")

    if rows == 0:
        return None

    color = iter(plt.cm.Set2(np.linspace(0, 1, rows)))
    wl = np.array(list(range(350, 901, 1)))
    fig = plt.figure(figsize=(10, 8))
    for meas in l1b:
        Ed = list_to_array(meas['ed.irradiance'])
        Lu = list_to_array(meas['lu.radiance'])
        Ld = list_to_array(meas['ld.radiance'])
        if Ed is not None and Lu is not None and Ld is not None:
            #plot by quality ? Currently only read l2 quality (but this is a nice top level check)
            c = next(color)
            plt.subplot(311)
            plt.plot(wl, Ed, c=c, label=meas['measurement.date'][10:16])
            plt.subplot(312)
            plt.plot(wl, Ld, c=c)
            plt.subplot(313)
            plt.plot(wl, Lu, c=c)
            station = meas['site.name']
    plt.subplot(311)
    plt.ylabel('Ed [W/(m2*nm)]')
    plt.xlim(wl[0], wl[-1])
    plt.title('{} at {} on {}'.format(instrument, station, day))
    plt.legend(loc='upper left', prop={'size': 8}, bbox_to_anchor=(1.01, 1))
    plt.subplot(312)
    plt.ylabel('Ld [W/(m2*nm*sr)]')
    plt.xlim(wl[0], wl[-1])
    plt.subplot(313)
    plt.ylabel('Lu [W/(m2*nm*sr)]')
    plt.xlim(wl[0], wl[-1])
    plt.xlabel('Wavelength [nm]')
    return fig



# -*- coding: utf-8 -*-
"""
Example/test downloading WISPstation data and optionally plotting Rrs and (ir)radiance curves and/or data download reports from WISPcloud

For full command line argument options see:
    python WISP_retrieval_plot_test.py -h

Example:
    python WISP_retrieval_plot_test.py -c -o -r -s '08:00:00' -e '16:00:00' -d '2019-06-21' -i 'WISPstation001' -t './test'


"""


from monda.WISP import plots
import os
import sys
import datetime
import logging
import argparse

log = logging.getLogger('WISP-test')
myFormat = '%(asctime)s | %(name)s | %(levelname)s | %(message)s'
formatter = logging.Formatter(myFormat)
logging.basicConfig(level = 'INFO', format = myFormat, stream = sys.stdout)


def run_example(instrument, date, start_time, end_time, target, output_radiance, output_rrs, output_data):
    log.info('running WISP test')

    if output_data:
        output_file_rrs = os.path.join(target, 'WISP_test_data_Rrs.txt')
        output_file_rad = os.path.join(target, 'WISP_test_data_Ed-Lt-Ls.txt')
    else:
        output_file_rrs = None
        output_file_rad = None

    plot1 = plots.plot_reflectance_station(instrument, date, start_time, end_time, output_file_rrs)
    plot2 = plots.plot_radiances_station(instrument, date, start_time, end_time, output_file_rad)

    if output_data:
        log.info(f"Rrs data saved to {output_file_rrs}")
        log.info(f"(ir)radiance data saved to {output_file_rad}")

    if plot1 is None:
       log.info("No data retrieved")
    elif output_rrs:
        plot_file_rrs = os.path.join(target, 'WISP_test_plot_Rrs.png')
        log.info(f"Saving Rrs plot to {plot_file_rrs}")
        plot1.savefig(plot_file_rrs, format='png', dpi=150)


    if plot1 is None:
       log.info("No data retrieved")
    elif output_radiance:
        plot_file_rad = os.path.join(target, 'WISP_test_plot_Ed-Lt-Ls.png')
        log.info(f"Saving (ir)radiances plot to {plot_file_rad}")
        plot2.savefig(plot_file_rad, format='png', dpi=150)


def parse_args():
    """Interpret command line arguments"""
    parser = argparse.ArgumentParser()
    parser.add_argument('-i','--instrument',  required = False, type = str, default = 'WISPstation001', help = "WISP Instrument to retrieve data from (default WISPstation001)")
    parser.add_argument('-d','--date',        required = False, type = str, default = "2019-06-21", help = "Date to retrieve observation data from in format 'YYYY-mm-dd'")
    parser.add_argument('-s','--start_time',  required = False, type = str, default = "09:00:00", help = "First UTC Time to include in data retrieval in format 'HH:MM:SS' (default 09:00:00)")
    parser.add_argument('-e','--end_time',    required = False, type = str, default = "15:00:00", help = "Last UTC time to include in data retrieval in format 'HH:MM:SS' (default 15:00:00)")

    parser.add_argument('-t','--target',      required = False, type = str, default=None, help = "Path to target folder for plots (defaults to 'WISP_testoutput' in the current folder).")

    parser.add_argument('-r','--output_radiance',required = False, action='store_true', help = "Save plot of Ls, Lt, Ed spectra")
    parser.add_argument('-c','--output_rrs',     required = False, action='store_true', help = "Save plot of Rrs spectra")
    parser.add_argument('-o','--output_data',    required = False, action='store_true', help = "Save data reports")

    args = parser.parse_args()

    return args


if __name__ == '__main__':

    args = parse_args()

    if not any([args.output_radiance, args.output_rrs]):
        log.warning("No plot outputs specified (see -h for help)")

    if args.target is None:
        args.target = os.path.join('.', 'WISP_test-output')

    if not os.path.isdir(args.target):
        os.mkdir(args.target)

    run_example(args.instrument, args.date, args.start_time, args.end_time,
                args.target, args.output_radiance, args.output_rrs, args.output_data)
